#!/bin/bash
########################################################################
# build_tauola.sh
#
# Build tauola and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_03_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"{64,}
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
package=tauola
origpkgver=v1_1_8
pkgver=${origpkgver}e # Append letters for local patches.
ssibuildshims_version=v2_04_01
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
####################################

##################
# Other useful variables. *C*
pkgtardir=TAUOLA
pkgtarfilestem=${package}-${pkgdotver}
pkgtarfile=${pkgtarfilestem}.tar.gz
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e17,e19,e20,e21,c2,c7} \
  --pqual={p2715a,p372,p383b} \
  --btype={debug,prof} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

if [[ "${pqual}" =~ ^p([[:digit:]]+) ]] && \
  { (( ${BASH_REMATCH[1]:0:1} == 2 )) || (( ${BASH_REMATCH[1]:1} < 83 )); } && \
  { { [ "${cqual:0:1}" = e ] && (( ${cqual#e} > 19 )); } || \
    { [ "${cqual:0:1}" = c ] && (( ${cqual#c} > 7 )); } || \
    ! [[ "${cqual:0:1}" == [ce] ]]; }; then
  case ${cqual:0:1} in
    e) emsg="GCC > 8.3";;
    c) emsg="LLVM/Clang > 7.0.0";;
    *) emsg="non-GCC / LLVM compilers"
  esac
  ssi_die "Python < 3.8.3 not supported with ${emsg}"
fi

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics IN_SOURCE TRACK_PROGRESS || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
#
# e.g.
#   setup_cmake --latest || exit
# OR
#   setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

####################################
# Patches.
#
# 1. Patch configure.in to not confuse UPS XXX_VERSION variables for LCG
#    ones, and not to forcibly override CXXFLAGS.
#
# 2. Remove overriding -std=c++11 flag from make.inc.in and allow
#    user-supplied CXXFLAGS and CPPFLAGS to override.
####################################
${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  -a "${pkgtardir}" \
  "${tardir}/${pkgtarfile}" \
  fix-configure-in.patch fix-make-inc-in.patch \
  || exit

##################
# TAUOLA C++ source distribution comes needing autoconf to be run. Look
# for a patch archive in patch/, or run autoconf and make one as part of
# this build process if we can.
autoconf_patch="${patchdir}/${pkgtarfilestem}-autoconf.tar.bz2"
[ -x gtar ] && tar=gtar || tar=tar
tarver="$(${tar} --version | head -1)"
if [[ "$tarver" == *GNU* ]]; then
  taropt=-N
elif [[ "$tarver" == bsdtar* ]]; then
  taropt=--newer-mtime-than
else
  ssi_die "cannot generate autoconf patch archive: no GNU tar and don't understand $tarver"
fi

if [ -f "${autoconf_patch}" ]; then
  ${trace} prepare_source "${ssi_args[@]}" "${autoconf_patch}" || exit
elif type -t autoconf >/dev/null 2>&1; then
  cd "${pkgtardir}" || ssi_die "unable to cd to ${sourcedir}/${pkgtardir}"
  ssi_info "running autoconf to generate patch archive ${autoconf_patch##*/}"
  ${trace} autoconf || \
    ssi_die "failed to run autoconf to generate ${autoconf_patch##*/}"
  cd - >/dev/null 2>&1
  ssi_info "archiving generated files to ${autoconf_patch##*/}"
  ${trace} tar jcf "${autoconf_patch}" ${taropt} \
    "${sourcedir}/${pkgtarfilestem}.stamp" "${pkgtardir}" || \
    ssi_die "unable to generate patch archive ${autoconf_patch}"
  ssi_info "regenerating source archive in ${product_dir}"
  ${trace} make_source_code_tarball "${ssi_args[@]}" \
    "${product_dir}" ${package} ${pkgver} || \
    ssi_die "unable to regenerate source archive after autoconf"
else
  ssi_die "missing patch archive ${autoconf_patch##*/} and no autoconf to generate"
fi
##################

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

declare -a cppflg cxxflg cflg fflg
case ${btype} in
  debug) fflg+=(-g -O0);;
  prof) fflg+=(-g -fno-omit-frame-pointer -O3 -DNDEBUG);;
  *) ssi_die -I "accepted unknown btype=${btype}"
esac
cflg+=("${fflg[@]}")
cxxflg+=("${fflg[@]}")
case ${cqual} in
  e1[79]|e20) cc=gcc; cxx=g++; cxxstd=17; fc=gfortran;;
  e21) cc=gcc; cxx=g++; cxxstd=2a; fc=gfortran;;
  c[27]) cc=clang; cxx=clang++; cxxstd=17; fc=gfortran;;
  *) ssi_die -I "accepted unknown cqual=${cqual}"
esac
cxxflg+=(-std=c++${cxxstd} -Wno-deprecated-declarations -Wno-unused-variable -Wno-unused-but-set-variable)
cflg+=(-std=c11)
fcflg+=(-std=legacy)
# Needed for a successful build of the legacy Fortran when --prefix is
# specified.
cppflg+=("-I${blddir}/${pkgtardir}/include")
case ${pqual} in
  p2715a) cxxflg+=(-Wno-register);;
  p372|p383b) :;;
  *) ssi_die -I "accepted unknown pqual=${pqual}"
esac

##################
# Configure parallelism. *C*
#ncores=$(ncores)
ncores=1 # Bad Makefiles, Bad!
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtardir} for in-source builds, ${blddir} otherwise.

# Current TAUOLA build system requires configure be invoked from the
# source (==build) directory.
${trace} cd "${blddir}/${pkgtardir}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*

${trace} autoconf && \
  ${trace} "${sourcedir}/${pkgtardir}/configure" --prefix="${pkgdir}" \
  --with-hepmc3="${HEPMC3_FQ_DIR}" \
  --without-hepmc \
  --with-lhapdf="${LHAPDF_FQ_DIR}" \
  --with-tau-spinner \
  CC=${cc} CFLAGS="${cppflg[*]} ${cflg[*]}" \
  CXX=${cxx} CXXFLAGS="${cppflg[*]} ${cxxflg[*]}" \
  F77=${fc} FFLAGS="${fcflg[*]}" \
  || ssi_die "configure failed"
##################

##################
# Build. *C*

${trace} make -j ${ncores} || ssi_die "build failed"
##################

##################
# Test. *C*
##################

##################
# Install. *C*

${trace} make -j ${ncores} install || ssi_die "install failed"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*

# Tidy source to install as debug source
${trace} make clean; find . -name '*~' | xargs rm -f

# Move cleaned source to installation point.
ssi_info "installing debug source"
${trace} cd "${pkgtop}" && \
  ${trace} rm -rf source && \
  ${trace} mv "${blddir}" source || \
  ssi_die "unable to move source for packaging"
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
